/*
 * Copyright 2014 Cloud Linux Zug GmbH
 *
 * Licensed under CLOUD LINUX LICENSE AGREEMENT
 * http://cloudlinux.com/docs/LICENSE.TXT
 *
 * The part the support library for lsapi & proxy_lsapi Apache modules
 * author Alexander Demeshko <ademeshko@cloudlinux.com>
 *
 */
 
#include <lscapi_util.h>

#include <util_script.h>
#include <http_request.h>

struct hdrs_context {
    const lscapi_resphdr_info_t *hdrsInfoPtr;
    int curNum;
    int curHdrSize;
    apr_off_t curOff;
    request_rec *r;
    lsapi_svr_conf_t *svrcfg;
};

static int getsfunc_HDRBUF(char *w, int len, void *phdrCon) {
    struct hdrs_context *c = (struct hdrs_context*)phdrCon;

    if(c->curNum >= c->hdrsInfoPtr->hdrNum) { // normal exit
        w[0] = '\0';
        return 1;
    }

    if(c->curOff > c->hdrsInfoPtr->dataLen) {
        if(c->svrcfg->debug_enabled) {
            lscapi_rlog(APLOG_NOTICE, 0, c->r,
                        "Wrong packet format: offset %" APR_OFF_T_FMT " is more than datalen %"APR_SIZE_T_FMT, 
                        c->curOff, c->hdrsInfoPtr->dataLen);
        }
        return -1;
    }

    if(!c->curHdrSize) {
        c->curHdrSize = c->hdrsInfoPtr->hdrSizes[c->curNum];
    }


    if(c->curHdrSize < len) {
        memcpy(w, c->hdrsInfoPtr->buf + c->curOff, c->curHdrSize);
        c->curNum++;
        c->curOff += c->curHdrSize;
        c->curHdrSize = 0;
    } else {
        memcpy(w, c->hdrsInfoPtr->buf + c->curOff, len-1);
        w[len-1] = '\0';
        c->curOff += len-1;
        c->curHdrSize -= len-1;
    }

    return 1;
}

#ifdef LSCAPI_WITH_DUMP_HEADER
/*
static void DumpHeaders(request_rec *r, const lscapi_resphdr_info_t *hdrsInfoPtr) {
    static unsigned long serialNo;
    char fname[256];
    
    
    snprintf(fname, sizeof fname, "/tmp/%s-%08lu.log", r->hostname, serialNo);
    serialNo++;

    FILE *fptr = fopen(fname, "wb");
    if(fptr) {
        lscapi_rlog(APLOG_ERR, 0, r,
                    "Dump headers into (%s)", fname);
        fwrite(hdrsInfoPtr->buf, 1, hdrsInfoPtr->dataLen, fptr);
        fclose(fptr);
    }
}
*/
#endif

static const char * get_status_line(int status)                                                                                                                         
{                                                                                                                                                                       
    return (status == HTTP_MOVED_TEMPORARILY) ? "302 Moved Temporarily" : ap_get_status_line(status);                                                                   
}

#define SET_STATUS_WITH_LINE(r, st)  do { (r)->status = (st); (r)->status_line = get_status_line((r)->status); } while(0) 

static int ParseHeaders(request_rec *r, lsapi_svr_conf_t *svrcfg, const lscapi_resphdr_info_t *hdrsInfoPtr) {
    char buffer[MAX_STRING_LEN];
    struct hdrs_context context;
    context.hdrsInfoPtr = hdrsInfoPtr;
    context.curHdrSize = 0;
    context.curOff = sizeof(lsapi_resp_info) + sizeof(uint16_t)*hdrsInfoPtr->hdrNum;
    context.curNum = 0;
    context.svrcfg = svrcfg;
    context.r = r;

    int rc = ap_scan_script_header_err_core(r, buffer, getsfunc_HDRBUF, (void *) &context);
    if(hdrsInfoPtr->respStatus) SET_STATUS_WITH_LINE(r, hdrsInfoPtr->respStatus);
    return rc;
}

static int read_from_backend(lsapi_conn_t *conn, lsapi_svr_conf_t *svrcfg, int drain, char *errbuf, size_t errlen) {
    int rc;
    request_rec *r = conn->r;
    char buf[LSAPI_MAX_DATA_PACKET_LEN];
    unsigned eventMask;
    apr_bucket_brigade *ob = apr_brigade_create(r->pool, r->connection->bucket_alloc);

    if(!drain) {
        ob = apr_brigade_create(r->pool, r->connection->bucket_alloc);
    }

    while(1) {

        size_t bufLen = LSAPI_MAX_DATA_PACKET_LEN;
        eventMask = 0;
        rc = lscapi_receive_response_chunk(conn->backend, buf, &bufLen, &eventMask, errbuf, errlen);
        if(rc != 0) {
            lscapi_rlog(APLOG_ERR, 0, r, "Error receiving response: %s", errbuf);
            lscapi_set_error(conn->lscapi); 
            // TODO: it is too late to change status here, as header is already processed. But what we have to do?
            if(!drain) {
                SET_STATUS_WITH_LINE(r, HTTP_INTERNAL_SERVER_ERROR);
            }
            return -1;
        }

        if(eventMask & LSCAPI_BACKEND_LOG_RECEIVED) {
            if(eventMask & LSCAPI_BACKEND_LOG_FATAL) {
                lscapi_rlog(APLOG_ERR, 0, r, "Backend fatal error: %s", lscapi_get_backend_log(conn->lscapi) );
            } else if(svrcfg->debug_enabled) {
                lscapi_rlog(APLOG_NOTICE, 0, r, "Backend log: %s", lscapi_get_backend_log(conn->lscapi) );
            }
        }
        
        if(eventMask & LSCAPI_RESPONSE_FINISHED) {
            if(!drain) {
                apr_bucket *eosB = apr_bucket_eos_create(r->connection->bucket_alloc);
                APR_BRIGADE_INSERT_TAIL(ob, eosB);
                ap_pass_brigade(r->output_filters, ob);
            }
            break;
        }

        if(!drain) {
            apr_bucket *b = apr_bucket_transient_create(buf, bufLen, r->connection->bucket_alloc);
            APR_BRIGADE_INSERT_TAIL(ob, b);
            ap_pass_brigade(r->output_filters, ob);
        }

    } //while(!done)

    if(!drain) {
        //apr_brigade_cleanup(ib);
        apr_brigade_cleanup(ob);
    }
    return 0;
}


apr_status_t lscapi_do_request(lsapi_conn_t *conn, lsapi_svr_conf_t *svrcfg)
{
    char errbuf[256];
    int rc;
    request_rec *r = conn->r;
    unsigned eventMask;
        
    eventMask = 0;
    rc = lscapi_send_request(conn->backend, &eventMask, errbuf, sizeof errbuf);
    if (rc != 0) {
        lscapi_rlog(APLOG_ERR, 0, r, "Error sending request (lsphp is killed?): %s", errbuf);
        lscapi_set_error(conn->lscapi);
        SET_STATUS_WITH_LINE(r, HTTP_INTERNAL_SERVER_ERROR);
        return r->status;
    }

    if(eventMask & LSCAPI_BACKEND_LOG_RECEIVED) {
        if(eventMask & LSCAPI_BACKEND_LOG_FATAL) {
            lscapi_rlog(APLOG_ERR, 0, r, "Backend fatal error: %s", lscapi_get_backend_log(conn->lscapi) );
        } else if(svrcfg->debug_enabled) {
            lscapi_rlog(APLOG_NOTICE, 0, r, "Backend log: %s", lscapi_get_backend_log(conn->lscapi) );
        }
    }

    lscapi_resphdr_info_t hdrsInfo;

    eventMask = 0;
    rc = lscapi_receive_response_header(conn->backend, &hdrsInfo, &eventMask, errbuf, sizeof errbuf);
    if(rc != 0) {
        lscapi_rlog(APLOG_ERR, 0, r, "Error receiving response header (lsphp is killed?): %s", errbuf);
        lscapi_set_error(conn->lscapi);
        SET_STATUS_WITH_LINE(r, HTTP_INTERNAL_SERVER_ERROR);
        return r->status;
    }

    if(eventMask & LSCAPI_BACKEND_LOG_RECEIVED) {
        if(eventMask & LSCAPI_BACKEND_LOG_FATAL) {
            lscapi_rlog(APLOG_ERR, 0, r, "Backend fatal error: %s", lscapi_get_backend_log(conn->lscapi) );
        } else if(svrcfg->debug_enabled) {
            lscapi_rlog(APLOG_NOTICE, 0, r, "Backend log: %s", lscapi_get_backend_log(conn->lscapi) );
        }
    }

    rc = ParseHeaders(r, svrcfg, &hdrsInfo);
    if (rc == HTTP_NOT_MODIFIED)
    {
        return rc;
    } else if (rc != APR_SUCCESS)
    {
        read_from_backend(conn, svrcfg, 1, errbuf, sizeof errbuf);  // discard output
        return HTTP_INTERNAL_SERVER_ERROR;
    }

    const char *location = apr_table_get(r->headers_out, "Location");
    if (location && location[0] == '/' && r->status == 200)
    {
        read_from_backend(conn, svrcfg, 1, errbuf, sizeof errbuf);  // discard output
        r->method = "GET";
        r->method_number = M_GET;
        apr_table_unset(r->headers_in, "Content-Length");

        ap_internal_redirect_handler(location, r);
        return OK;
    }
    else if (location && r->status == 200)
    {
        read_from_backend(conn, svrcfg, 1, errbuf, sizeof errbuf);  // discard output
        return HTTP_MOVED_TEMPORARILY;
    }

    rc = read_from_backend(conn, svrcfg, 0, errbuf, sizeof errbuf); // send output to browser through filters
    if(rc != 0) {
        return r->status;
    }

    return OK;
}
